import copy
import enum
import functools
import json
import pathlib
import threading
import typing

import calibration_client
import h5py
import numpy as np
from calibration_client.modules import (
    Calibration,
    CalibrationConstant,
    CalibrationConstantVersion,
    Detector,
    DetectorType,
    Parameter,
    PhysicalDetectorUnit,
)
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
)

from . import utils


class ConditionNotFound(Exception):
    pass


class DetectorNotFound(Exception):
    pass


class ModuleNotFound(Exception):
    pass


class CalibrationNotFound(Exception):
    pass


class CalibrationClientConfigError(Exception):
    pass


def _check_resp(resp, exception=Exception):
    # TODO: probably verify using "info" that exception is the right one
    if not resp["success"]:
        # TODO: probably more types of app_info errors?
        if resp["app_info"]:
            if "not found" in resp["info"]:
                # this was likely the exception exception
                raise exception(resp["info"])
            else:
                # but could also be authorization or similar issue
                raise CalibrationClientConfigError(resp["app_info"])
        raise exception(resp["info"])


def _add_status_schema_from_enum(schema, prefix, enum_class):
    for constant in enum_class:
        constant_node = f"{prefix}.{constant.name}"
        (
            NODE_ELEMENT(schema).key(constant_node).commit(),
            BOOL_ELEMENT(schema)
            .key(f"{constant_node}.found")
            .readOnly()
            .initialValue(False)
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.createdAt")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.calibrationId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.conditionId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.constantId")
            .readOnly()
            .initialValue("")
            .commit(),
        )


class DetectorStandin(typing.NamedTuple):
    detector_name: str
    modno_to_source: dict
    frames_per_train: int
    module_shape: tuple


class OperatingConditions(dict):
    # TODO: support deviation?
    def encode(self):
        return {
            "parameters": [
                {
                    "parameter_name": key,
                    "lower_deviation_value": 0.0,
                    "upper_deviation_value": 0.0,
                    "flg_available": False,
                    "value": value,
                }
                for (key, value) in self.items()
            ]
        }

    def __hash__(self):
        # this takes me back to pre-screening interview time...
        return hash(tuple(sorted(self.items())))


class BaseCalcatFriend:
    _constant_enum_class = None  # subclass should set
    _constants_need_conditions = None  # subclass should set

    @staticmethod
    def add_schema(
        schema,
        detector_type,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        """Add elements needed by this object to device's schema (expectedSchema)

        All elements added to schema go under prefixes which should end with name of
        node which does not exist yet. To change default values and add more fields,
        extend this method in subclass.

        The param_prefix node will hold all the parameters needed to build constant
        condition dicts for querying CalCat. These values are set either directly on
        the device or via manager and this class gets them from the device using helper
        function _get_param. See for example AgipdCalcatFriend.dark_condition.

        The status_prefix node is used to report information about what was found in
        CalCat. This class will update the values on the device using the helper
        function _set_status. This should not need to happen in subclass methods.
        """

        (
            NODE_ELEMENT(schema)
            .key(param_prefix)
            .displayedName("Constant retrieval parameters")
            .commit(),
            NODE_ELEMENT(schema)
            .key(status_prefix)
            .displayedName("Constants retrieved")
            .commit(),
        )

        # Parameters which any detector would probably have (extend this in subclass)
        # TODO: probably switch to floating point for everything, including mem cells
        (
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorType")
            .displayedName("Detector type name")
            .description(
                "Name of detector type in CalCat; typically has suffix '-Type'"
            )
            .readOnly()
            .initialValue(detector_type)
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorName")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.karaboDa")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.memoryCells")
            .assignmentOptional()
            .defaultValue(352)
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsX")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsY")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.biasVoltage")
            .assignmentOptional()
            .defaultValue(300)
            .reconfigurable()
            .commit(),
        )

    def __init__(
        self,
        device,
        secrets_fn: pathlib.Path,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        self.device = device
        self.param_prefix = param_prefix
        self.status_prefix = status_prefix
        self.cached_constants = {}

        if not secrets_fn.is_file():
            self.device.log.WARN(f"Missing CalCat secrets file (expected {secrets_fn})")
        with secrets_fn.open("r") as fd:
            calcat_secrets = json.load(fd)

        self.caldb_store = pathlib.Path(calcat_secrets["caldb_store_path"])
        if not self.caldb_store.is_dir():
            raise ValueError(f"caldb_store location '{self.caldb_store}' is not dir")

        self.device.log.INFO(f"Connecting to CalCat at {calcat_secrets['base_url']}")
        base_url = calcat_secrets["base_url"]
        self.client = calibration_client.CalibrationClient(
            client_id=calcat_secrets["client_id"],
            client_secret=calcat_secrets["client_secret"],
            user_email=calcat_secrets["user_email"],
            base_api_url=f"{base_url}/api/",
            token_url=f"{base_url}/oauth/token",
            refresh_url=f"{base_url}/oauth/token",
            auth_url=f"{base_url}/oauth/authorize",
            scope="public",
            session_token=None,
        )
        self.device.log.INFO("CalCat connection established")

    def _get_param(self, key):
        """Helper to get value from attached device schema"""
        return self.device.get(f"{self.param_prefix}.{key}")

    def _set_status(self, key, value):
        """Helper to update information about found constants on device"""
        self.device.set(f"{self.status_prefix}.{key}", value)

    @functools.cached_property
    def detector_id(self):
        resp = Detector.get_by_identifier(self.client, self._get_param("detectorName"))
        _check_resp(resp, DetectorNotFound)
        return resp["data"]["id"]

    @functools.cached_property
    def detector_type_id(self):
        resp = DetectorType.get_by_name(self.client, self._get_param("detectorType"))
        _check_resp(resp, DetectorNotFound)
        return resp["data"]["id"]

    # TODO: support updating mapping maybe?
    # (just means del self.pdus and the properties depending on it)
    @functools.cached_property
    def pdus(self):
        resp = PhysicalDetectorUnit.get_all_by_detector_id(
            self.client, self.detector_id
        )
        _check_resp(resp)
        for irrelevant_key in ("detector", "detector_type", "flg_available"):
            for pdu in resp["data"]:
                del pdu[irrelevant_key]
        return resp["data"]

    @functools.cached_property
    def _karabo_da_to_float_uuid(self):
        return {pdu["karabo_da"]: pdu["float_uuid"] for pdu in self.pdus}

    @functools.cached_property
    def _karabo_da_to_id(self):
        return {pdu["karabo_da"]: pdu["id"] for pdu in self.pdus}

    @utils.threadsafe_cache
    def calibration_id(self, calibration_name: str):
        resp = Calibration.get_by_name(self.client, calibration_name)
        # TODO: dump name in exception
        _check_resp(resp, CalibrationNotFound)
        return resp["data"]["id"]

    @utils.threadsafe_cache
    def condition_id(self, pdu, condition):
        # modifying condition parameter messes with cache
        condition_with_detector = copy.copy(condition)
        condition_with_detector["Detector UUID"] = pdu
        self.device.log.INFO(f"Look for condition: {condition_with_detector}")
        resp = self.client.search_possible_conditions_from_dict(
            "", condition_with_detector.encode()
        )
        _check_resp(resp, ConditionNotFound)
        return resp["data"][0]["id"]

    @utils.threadsafe_cache
    def constant_id(self, calibration_id, condition_id):
        resp = CalibrationConstant.get_by_uk(
            self.client,
            calibration_id=calibration_id,
            detector_type_id=self.detector_type_id,
            condition_id=condition_id,
        )
        _check_resp(resp)
        return resp["data"]["id"]

    def get_constant_version(self, constant, snapshot_at=None):
        # TODO: support snapshot
        # TODO: catch exceptions, give warnings appropriately
        karabo_da = self._get_param("karaboDa")
        self.device.log.DEBUG(f"Will look for {constant} for {karabo_da}")

        if karabo_da not in self._karabo_da_to_float_uuid:
            raise ModuleNotFound(f"Module map did not include {karabo_da}")

        if isinstance(constant, str):
            constant = self._constant_enum_class[constant]

        calibration_id = self.calibration_id(constant.name)
        self._set_status(f"{constant.name}.calibrationId", calibration_id)

        condition = self._constants_need_conditions[constant]()
        condition_id = self.condition_id(
            self._karabo_da_to_float_uuid[karabo_da], condition
        )
        self._set_status(f"{constant.name}.conditionId", condition_id)

        constant_id = self.constant_id(
            calibration_id=calibration_id, condition_id=condition_id
        )
        self._set_status(f"{constant.name}.constantId", constant_id)

        resp = CalibrationConstantVersion.get_by_uk(
            self.client,
            calibration_constant_id=constant_id,
            physical_detector_unit_id=self._karabo_da_to_id[karabo_da],
            event_at=None,
            snapshot_at=None,
        )
        _check_resp(resp)
        timestamp = resp["data"][
            "begin_at"
        ]  # TODO: check which key we like (also has begin_validity_at)
        self._set_status(f"{constant.name}.createdAt", timestamp)

        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        # TODO: handle FileNotFoundError if we are led astray
        with h5py.File(file_path, "r") as fd:
            constant_data = np.array(fd[resp["data"]["data_set_name"]]["data"])
        self.cached_constants[constant] = constant_data
        self._set_status(f"{constant.name}.found", True)

        return constant_data

    def get_constant_version_and_call_me_back(
        self, constant, callback, snapshot_at=None
    ):
        """WIP. Callback function will get constant name, constant data, and hopefully
        soon also some metadata or whatever.
        """
        # TODO: do we want to use asyncio / "modern" async?
        def aux():
            data = self.get_constant_version(constant, snapshot_at)
            callback(constant, data)

        thread = threading.Thread(target=aux)
        thread.start()
        return thread

    def flush_constants(self):
        for constant in self._constant_enum_class:
            self._set_status(f"{constant.name}.createdAt", "")
            self._set_status(f"{constant.name}.found", False)


class AgipdConstants(enum.Enum):
    SlopesFF = enum.auto()
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesPC = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


class AgipdCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = AgipdConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            AgipdConstants.ThresholdsDark: self.dark_condition,
            AgipdConstants.Offset: self.dark_condition,
            AgipdConstants.SlopesPC: self.dark_condition,
            AgipdConstants.SlopesFF: self.illuminated_condition,
            AgipdConstants.BadPixelsDark: self.dark_condition,
            AgipdConstants.BadPixelsPC: self.dark_condition,
            AgipdConstants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(
        schema, param_prefix="constantParameters", status_prefix="foundConstants"
    ):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(
            schema, "AGIPD-Type", param_prefix, status_prefix
        )

        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.gainSetting")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.gainMode")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.integrationtime")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )

        _add_status_schema_from_enum(schema, status_prefix, AgipdConstants)

    def dark_condition(self):
        res = OperatingConditions()
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Memory cells"] = self._get_param("memoryCells")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        return res

    def illuminated_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Source Energy"] = self._get_param("photonEnergy")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        res["Gain Setting"] = self._get_param("gainSetting")
        return res


class DsscConstants(enum.Enum):
    Offset = enum.auto()


class DsscCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = DsscConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            DsscConstants.Offset: self.dark_condition,
        }

    @staticmethod
    def add_schema(
        schema, param_prefix="constantParameters", status_prefix="foundConstants"
    ):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(
            schema, "DSSC-Type", param_prefix, status_prefix
        )
        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.pulseIdChecksum")
            .assignmentOptional()
            .defaultValue(2.8866323107820637e-36)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(4.5)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.encodedGain")
            .assignmentOptional()
            .defaultValue(67328)
            .commit(),
        )

        _add_status_schema_from_enum(schema, status_prefix, DsscConstants)

    def dark_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        # res["Pulse id checksum"] = self._get_param("pulseIdChecksum")
        # res["Acquisition rate"] = self._get_param("acquisitionRate")
        # res["Encoded gain"] = self._get_param("encodedGain")
        return res
